@@ -14,22 +14,29 @@
 #define GETNEXT getchar
 
 enum JSONType {
-	TYPE_UNKNOWN = 0,
-	TYPE_PRIMITIVE,
-	TYPE_STRING,
-	TYPE_ARRAY,
-	TYPE_OBJECT
+	TYPE_PRIMITIVE = 'p',
+	TYPE_STRING    = 's',
+	TYPE_ARRAY     = 'a',
+	TYPE_OBJECT    = 'o'
 };
 
-#define JSON_MAX_NODE_DEPTH 32
+#define JSON_MAX_NODE_DEPTH 64
 
 struct json_node {
 	enum JSONType type;
 	char *name;
 	size_t namesiz;
-	size_t index; /* count/index for TYPE_ARRAY and TYPE_OBJECT */
+	size_t index; /* count/index for array or object type */
 };
 
+const char *JSON_ERROR_ALLOC         = "cannot allocate enough memory";
+const char *JSON_ERROR_BALANCE       = "unbalanced nodes";
+const char *JSON_ERROR_CODEPOINT     = "invalid codepoint";
+const char *JSON_ERROR_DEPTH         = "max node depth reached";
+const char *JSON_ERROR_ESCAPE_CHAR   = "unknown escape character in string";
+const char *JSON_ERROR_INVALID_CHAR  = "invalid character in string";
+const char *JSON_ERROR_OBJECT_MEMBER = "object member, but not in an object";
+
 static int showindices = 0; /* -n flag: show indices count for arrays */
 
 int
@@ -80,7 +87,7 @@ capacity(char **value, size_t *sz, size_t cur, size_t inc)
 	size_t need, newsiz;
 	char *newp;
 
-	/* check addition overflow */
+	/* check for addition overflow */
 	if (cur > SIZE_MAX - inc) {
 		errno = EOVERFLOW;
 		return -1;
@@ -88,10 +95,10 @@ capacity(char **value, size_t *sz, size_t cur, size_t inc)
 	need = cur + inc;
 
 	if (need > *sz) {
-		if (need > SIZE_MAX - 16384) {
+		if (need > SIZE_MAX / 2) {
 			newsiz = SIZE_MAX;
 		} else {
-			for (newsiz = *sz; newsiz < need; newsiz += 16384)
+			for (newsiz = *sz < 64 ? 64 : *sz; newsiz <= need; newsiz *= 2)
 				;
 		}
 		if (!(newp = realloc(*value, newsiz)))
@@ -107,46 +114,53 @@ parsejson(void (*cb)(struct json_node *, size_t, const char *), const char **err
 {
 	struct json_node nodes[JSON_MAX_NODE_DEPTH] = { 0 };
 	size_t depth = 0, v = 0, vz = 0;
-	long cp;
+	long cp, hi, lo;
 	int c, i, escape, ret = -1;
 	char *value = NULL;
 
-	*errstr = "cannot allocate enough memory";
+	*errstr = JSON_ERROR_ALLOC;
 	if (capacity(&(nodes[0].name), &(nodes[0].namesiz), 0, 1) == -1)
 		goto end;
 	nodes[0].name[0] = '\0';
+	nodes[depth].type = TYPE_PRIMITIVE;
 
 	while ((c = GETNEXT()) != EOF) {
-		/* not whitespace or control-character */
+		/* not whitespace or control character */
 		if (c <= 0x20 || c == 0x7f)
 			continue;
 
 		switch (c) {
 		case ':':
-			nodes[depth].type = TYPE_PRIMITIVE;
-			if (v) {
-				if (!depth || nodes[depth - 1].type != TYPE_OBJECT) {
-					*errstr = "object member, but not in an object";
-					goto end;
-				}
-				value[v] = '\0';
-				if (capacity(&(nodes[depth].name), &(nodes[depth].namesiz), v, 1) == -1)
-					goto end;
-				memcpy(nodes[depth].name, value, v);
-				nodes[depth].name[v] = '\0';
-				v = 0;
+			if (!depth || nodes[depth - 1].type != TYPE_OBJECT) {
+				*errstr = JSON_ERROR_OBJECT_MEMBER;
+				goto end;
 			}
+
+			if (capacity(&value, &vz, v, 1) == -1)
+				goto end;
+			value[v] = '\0';
+			if (capacity(&(nodes[depth].name), &(nodes[depth].namesiz), v, 1) == -1)
+				goto end;
+			memcpy(nodes[depth].name, value, v);
+			nodes[depth].name[v] = '\0';
+			v = 0;
+			nodes[depth].type = TYPE_PRIMITIVE;
 			break;
 		case '"':
-			v = 0;
 			nodes[depth].type = TYPE_STRING;
-			for (escape = 0; (c = GETNEXT()) != EOF;) {
-				if (iscntrl(c))
-					continue;
+			escape = 0;
+			for (;;) {
+				c = GETNEXT();
+chr:
+				if (c < 0x20) {
+					/* EOF or control char: 0x7f is not defined as a control char in RFC8259 */
+					*errstr = JSON_ERROR_INVALID_CHAR;
+					goto end;
+				}
 
 				if (escape) {
+escchr:
 					escape = 0;
-
 					switch (c) {
 					case '"': /* FALLTHROUGH */
 					case '\\':
@@ -157,21 +171,52 @@ parsejson(void (*cb)(struct json_node *, size_t, const char *), const char **err
 					case 'r': c = '\r'; break;
 					case 't': c = '\t'; break;
 					case 'u': /* hex hex hex hex */
+						if (capacity(&value, &vz, v, 4) == -1)
+							goto end;
 						for (i = 12, cp = 0; i >= 0; i -= 4) {
 							if ((c = GETNEXT()) == EOF || !isxdigit(c)) {
-								*errstr = "invalid codepoint";
+								*errstr = JSON_ERROR_CODEPOINT;
 								goto end;
 							}
 							cp |= (hexdigit(c) << i);
 						}
-						if (capacity(&value, &vz, v, 5) == -1)
-							goto end;
+						/* RFC8259 - 7. Strings - surrogates.
+						 * 0xd800 - 0xdb7f - high surrogates */
+						if (cp >= 0xd800 && cp <= 0xdb7f) {
+							if ((c = GETNEXT()) != '\\') {
+								v += codepointtoutf8(cp, &value[v]);
+								goto chr;
+							}
+							if ((c = GETNEXT()) != 'u') {
+								v += codepointtoutf8(cp, &value[v]);
+								goto escchr;
+							}
+							for (hi = cp, i = 12, lo = 0; i >= 0; i -= 4) {
+								if ((c = GETNEXT()) == EOF || !isxdigit(c)) {
+									*errstr = JSON_ERROR_CODEPOINT;
+									goto end;
+								}
+								lo |= (hexdigit(c) << i);
+							}
+							/* 0xdc00 - 0xdfff - low surrogates */
+							if (lo >= 0xdc00 && lo <= 0xdfff) {
+								cp = (hi << 10) + lo - 56613888; /* - offset */
+							} else {
+								/* handle graceful: raw invalid output bytes */
+								v += codepointtoutf8(hi, &value[v]);
+								if (capacity(&value, &vz, v, 4) == -1)
+									goto end;
+								v += codepointtoutf8(lo, &value[v]);
+								continue;
+							}
+						}
 						v += codepointtoutf8(cp, &value[v]);
 						continue;
 					default:
-						continue; /* ignore unknown escape char */
+						*errstr = JSON_ERROR_ESCAPE_CHAR;
+						goto end;
 					}
-					if (capacity(&value, &vz, v, 2) == -1)
+					if (capacity(&value, &vz, v, 1) == -1)
 						goto end;
 					value[v++] = c;
 				} else if (c == '\\') {
@@ -179,7 +224,7 @@ parsejson(void (*cb)(struct json_node *, size_t, const char *), const char **err
 				} else if (c == '"') {
 					break;
 				} else {
-					if (capacity(&value, &vz, v, 2) == -1)
+					if (capacity(&value, &vz, v, 1) == -1)
 						goto end;
 					value[v++] = c;
 				}
@@ -191,7 +236,7 @@ parsejson(void (*cb)(struct json_node *, size_t, const char *), const char **err
 		case '[':
 		case '{':
 			if (depth + 1 >= JSON_MAX_NODE_DEPTH) {
-				*errstr = "max node depth reached";
+				*errstr = JSON_ERROR_DEPTH;
 				goto end;
 			}
 
@@ -203,25 +248,25 @@ parsejson(void (*cb)(struct json_node *, size_t, const char *), const char **err
 
 			depth++;
 			nodes[depth].index = 0;
+			nodes[depth].type = TYPE_PRIMITIVE;
 			if (capacity(&(nodes[depth].name), &(nodes[depth].namesiz), v, 1) == -1)
 				goto end;
 			nodes[depth].name[0] = '\0';
-			nodes[depth].type = TYPE_PRIMITIVE;
 			break;
 		case ']':
 		case '}':
 		case ',':
-			if (v &&
-			    (nodes[depth].type == TYPE_STRING ||
-			    nodes[depth].type == TYPE_PRIMITIVE)) {
+			if (v || nodes[depth].type == TYPE_STRING) {
+				if (capacity(&value, &vz, v, 1) == -1)
+					goto end;
 				value[v] = '\0';
 				cb(nodes, depth + 1, value);
 				v = 0;
 			}
 			if (!depth ||
-			    (nodes[depth - 1].type == TYPE_OBJECT && c == ']') ||
-			    (nodes[depth - 1].type == TYPE_ARRAY && c == '}')) {
-				*errstr = "unbalanced nodes";
+			    (c == ']' && nodes[depth - 1].type != TYPE_ARRAY) ||
+			    (c == '}' && nodes[depth - 1].type != TYPE_OBJECT)) {
+				*errstr = JSON_ERROR_BALANCE;
 				goto end;
 			}
 
@@ -233,11 +278,21 @@ parsejson(void (*cb)(struct json_node *, size_t, const char *), const char **err
 			}
 			break;
 		default:
-			if (capacity(&value, &vz, v, 2) == -1)
+			if (capacity(&value, &vz, v, 1) == -1)
 				goto end;
 			value[v++] = c;
 		}
 	}
+	if (depth) {
+		*errstr = JSON_ERROR_BALANCE;
+		goto end;
+	}
+	if (v || nodes[depth].type == TYPE_STRING) {
+		if (capacity(&value, &vz, v, 1) == -1)
+			goto end;
+		value[v] = '\0';
+		cb(nodes, depth + 1, value);
+	}
 
 	ret = 0; /* success */
 	*errstr = NULL;
@@ -292,13 +347,7 @@ processnode(struct json_node *nodes, size_t depth, const char *value)
 	}
 
 	putchar('\t');
-	switch (nodes[depth - 1].type) {
-	case TYPE_UNKNOWN:   return;
-	case TYPE_ARRAY:     putchar('a'); break;
-	case TYPE_OBJECT:    putchar('o'); break;
-	case TYPE_PRIMITIVE: putchar('p'); break;
-	case TYPE_STRING:    putchar('s'); break;
-	}
+	putchar(nodes[depth - 1].type);
 	putchar('\t');
 	printvalue(value);
 	putchar('\n');
